import torch
from torch import nn
import torch.nn.functional as F
from typing import Dict


class MultiOutputUnet3D(nn.Module):
    """
    3D U-Net architecture supporting multiple output heads (e.g., segmentation, flow).
    Adapted from Li, X. et al. Real-time denoising enables high-sensitivity fluorescence time-lapse imaging
    beyond the shot-noise limit. Nat Biotechnol 41, 282–292 (2023).
    """
    def __init__(
        self,
        in_channels: int = 1,
        output_heads: Dict[str, dict] = None,
        n_filter: int = 16,
        use_interpolation: bool = True
    ):
        """
        Parameters
        ----------
        in_channels : int
            Number of input channels (e.g. 1 for grayscale).
        output_heads : Dict[str, dict]
            Dictionary defining output heads, e.g.
            {
                'seg':  {'channels': 1, 'activation': 'sigmoid'},
                'flow': {'channels': 2, 'activation': None}
            }
        n_filter : int
            Base number of convolutional filters.
        use_interpolation : bool
            Whether to use interpolation instead of transposed convolutions.
        """
        super().__init__()
        self.output_heads = output_heads or {
            'default': {'channels': 1, 'activation': 'sigmoid'}
        }
        self.use_interpolation = use_interpolation

        # Encoding path
        self.encode1 = self.conv3d(in_channels, n_filter // 2)
        self.encode2 = self.conv3d(n_filter // 2, n_filter)
        if not use_interpolation:
            self.maxpool1 = nn.MaxPool3d(kernel_size=2, stride=2)

        self.encode3 = self.conv3d(n_filter, n_filter)
        self.encode4 = self.conv3d(n_filter, 2 * n_filter)
        if not use_interpolation:
            self.maxpool2 = nn.MaxPool3d(kernel_size=2, stride=2)

        self.encode5 = self.conv3d(2 * n_filter, 2 * n_filter)
        self.encode6 = self.conv3d(2 * n_filter, 4 * n_filter)
        if not use_interpolation:
            self.maxpool3 = nn.MaxPool3d(kernel_size=2, stride=2)

        # Middle
        self.middle_conv1 = self.conv3d(4 * n_filter, 4 * n_filter)
        self.middle_conv2 = self.conv3d(4 * n_filter, 8 * n_filter)

        # Decoding path
        if not use_interpolation:
            self.up1 = nn.ConvTranspose3d(8 * n_filter, 8 * n_filter, kernel_size=2, stride=2)
            self.up2 = nn.ConvTranspose3d(4 * n_filter, 4 * n_filter, kernel_size=2, stride=2)
            self.up3 = nn.ConvTranspose3d(2 * n_filter, 2 * n_filter, kernel_size=2, stride=2)
        else:
            self.up1_conv = self.conv3d(8 * n_filter, 8 * n_filter)
            self.up2_conv = self.conv3d(4 * n_filter, 4 * n_filter)
            self.up3_conv = self.conv3d(2 * n_filter, 2 * n_filter)

        self.decode1 = self.conv3d(12 * n_filter, 4 * n_filter)
        self.decode2 = self.conv3d(4 * n_filter, 4 * n_filter)
        self.decode3 = self.conv3d(6 * n_filter, 2 * n_filter)
        self.decode4 = self.conv3d(2 * n_filter, 2 * n_filter)
        self.decode5 = self.conv3d(3 * n_filter, n_filter)
        self.decode6 = self.conv3d(n_filter, n_filter // 2)

        # Output heads
        self.output_layers = nn.ModuleDict()
        for name, cfg in self.output_heads.items():
            self.output_layers[name] = nn.Conv3d(n_filter // 2, cfg['channels'], kernel_size=1)

    def conv3d(self, in_channels, out_channels, kernel_size=3, dropout=0.0, dilation=1):
        layers = [
            nn.Conv3d(in_channels, out_channels, kernel_size=kernel_size,
                      padding=dilation, dilation=dilation),
            nn.BatchNorm3d(out_channels),
            nn.LeakyReLU(negative_slope=0.1, inplace=True),
            nn.Dropout3d(dropout),
        ]
        return nn.Sequential(*layers)

    def concat(self, x1, x2):
        return torch.cat((x1, x2), dim=1)

    def apply_activation(self, x, activation):
        if activation == 'sigmoid':
            return torch.sigmoid(x)
        elif activation == 'tanh':
            return torch.tanh(x)
        elif activation == 'relu':
            return F.relu(x)
        return x

    def forward(self, x):
        # Encoding
        e1 = self.encode1(x)
        e2 = self.encode2(e1)

        if self.use_interpolation:
            m1 = F.interpolate(e2, scale_factor=0.5, mode='nearest')
        else:
            m1 = self.maxpool1(e2)

        e3 = self.encode3(m1)
        e4 = self.encode4(e3)

        if self.use_interpolation:
            m2 = F.interpolate(e4, scale_factor=0.5, mode='nearest')
        else:
            m2 = self.maxpool2(e4)

        e5 = self.encode5(m2)
        e6 = self.encode6(e5)

        if self.use_interpolation:
            m3 = F.interpolate(e6, scale_factor=0.5, mode='nearest')
        else:
            m3 = self.maxpool3(e6)

        # Middle
        mid1 = self.middle_conv1(m3)
        mid2 = self.middle_conv2(mid1)

        # Decoding
        if self.use_interpolation:
            u1 = F.interpolate(mid2, scale_factor=2, mode='nearest')
            u1 = self.up1_conv(u1)
        else:
            u1 = self.up1(mid2)
        c1 = self.concat(u1, e6)
        d1 = self.decode1(c1)
        d2 = self.decode2(d1)

        if self.use_interpolation:
            u2 = F.interpolate(d2, scale_factor=2, mode='nearest')
            u2 = self.up2_conv(u2)
        else:
            u2 = self.up2(d2)
        c2 = self.concat(u2, e4)
        d3 = self.decode3(c2)
        d4 = self.decode4(d3)

        if self.use_interpolation:
            u3 = F.interpolate(d4, scale_factor=2, mode='nearest')
            u3 = self.up3_conv(u3)
        else:
            u3 = self.up3(d4)
        c3 = self.concat(u3, e2)
        d5 = self.decode5(c3)
        d6 = self.decode6(d5)

        # Output heads
        outputs = {}
        for name, cfg in self.output_heads.items():
            logits = self.output_layers[name](d6)
            outputs[name] = self.apply_activation(logits, cfg.get('activation'))

        return outputs



from typing import Dict, Tuple, Union

import torch
from torch import nn


class FirstVGGBlock(nn.Module):
    """VGG block with Instance Normalization for the first layer"""

    def __init__(self, in_channels, middle_channels, out_channels, dropout=0.):
        super().__init__()
        self.relu = nn.LeakyReLU(negative_slope=0.1, inplace=True)
        self.conv1 = nn.Conv2d(in_channels, middle_channels, 3, padding=1)
        self.in1 = nn.InstanceNorm2d(middle_channels)  # Instance Norm instead of Batch Norm
        self.conv2 = nn.Conv2d(middle_channels, out_channels, 3, padding=1)
        self.in2 = nn.InstanceNorm2d(out_channels)  # Instance Norm instead of Batch Norm
        self.dropout = nn.Dropout2d(dropout)

    def forward(self, x):
        out = self.conv1(x)
        out = self.in1(out)
        out = self.relu(out)
        out = self.dropout(out)

        out = self.conv2(out)
        out = self.in2(out)
        out = self.relu(out)
        out = self.dropout(out)

        return out


class VGGBlock(nn.Module):
    def __init__(self, in_channels, middle_channels, out_channels, dropout=0., dilation=1):
        super().__init__()
        padding = dilation  # ensures same spatial dimensions
        self.relu = nn.LeakyReLU(negative_slope=0.1, inplace=True)
        self.conv1 = nn.Conv2d(in_channels, middle_channels, kernel_size=3, padding=padding, dilation=dilation)
        self.bn1 = nn.BatchNorm2d(middle_channels)
        self.conv2 = nn.Conv2d(middle_channels, out_channels, kernel_size=3, padding=padding, dilation=dilation)
        self.bn2 = nn.BatchNorm2d(out_channels)
        self.dropout = nn.Dropout2d(dropout)

    def forward(self, x):
        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)
        out = self.dropout(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)
        out = self.dropout(out)

        return out


class MultiOutputNestedUNet(nn.Module):
    def __init__(self, in_channels=1, output_heads: Dict[str, dict] = None, n_filter: int = 32,
                 deep_supervision: bool = False, dilation: Union[bool, Tuple[int, int, int, int, int]] = False,
                 train_mode: bool = True):
        super().__init__()
        self.output_heads = output_heads or {'default': {'channels': 1, 'activation': 'sigmoid'}}
        self.deep_supervision = deep_supervision
        self.train_mode = train_mode
        self.dilation = dilation
        if self.dilation is False:
            self.dilation = (1, 1, 1, 1, 1)

        nb_filter = [n_filter, n_filter * 2, n_filter * 4, n_filter * 8, n_filter * 16]

        self.pool = nn.MaxPool2d(2, 2)
        self.up = nn.Upsample(scale_factor=2, mode='bilinear', align_corners=True)

        self.conv0_0 = VGGBlock(in_channels, nb_filter[0], nb_filter[0], dilation=self.dilation[0])
        self.conv1_0 = VGGBlock(nb_filter[0], nb_filter[1], nb_filter[1], dilation=self.dilation[1])
        self.conv2_0 = VGGBlock(nb_filter[1], nb_filter[2], nb_filter[2], dilation=self.dilation[2])
        self.conv3_0 = VGGBlock(nb_filter[2], nb_filter[3], nb_filter[3], dilation=self.dilation[3])
        self.conv4_0 = VGGBlock(nb_filter[3], nb_filter[4], nb_filter[4], dilation=self.dilation[4])

        self.conv0_1 = VGGBlock(nb_filter[0] + nb_filter[1], nb_filter[0], nb_filter[0])
        self.conv1_1 = VGGBlock(nb_filter[1] + nb_filter[2], nb_filter[1], nb_filter[1])
        self.conv2_1 = VGGBlock(nb_filter[2] + nb_filter[3], nb_filter[2], nb_filter[2])
        self.conv3_1 = VGGBlock(nb_filter[3] + nb_filter[4], nb_filter[3], nb_filter[3])

        self.conv0_2 = VGGBlock(nb_filter[0] * 2 + nb_filter[1], nb_filter[0], nb_filter[0])
        self.conv1_2 = VGGBlock(nb_filter[1] * 2 + nb_filter[2], nb_filter[1], nb_filter[1])
        self.conv2_2 = VGGBlock(nb_filter[2] * 2 + nb_filter[3], nb_filter[2], nb_filter[2])

        self.conv0_3 = VGGBlock(nb_filter[0] * 3 + nb_filter[1], nb_filter[0], nb_filter[0])
        self.conv1_3 = VGGBlock(nb_filter[1] * 3 + nb_filter[2], nb_filter[1], nb_filter[1])

        self.conv0_4 = VGGBlock(nb_filter[0] * 4 + nb_filter[1], nb_filter[0], nb_filter[0])

        if self.deep_supervision:
            self.output_layers = nn.ModuleDict()
            for name, config in self.output_heads.items():
                self.output_layers[f"{name}_1"] = nn.Conv2d(nb_filter[0], config['channels'], kernel_size=1)
                self.output_layers[f"{name}_2"] = nn.Conv2d(nb_filter[0], config['channels'], kernel_size=1)
                self.output_layers[f"{name}_3"] = nn.Conv2d(nb_filter[0], config['channels'], kernel_size=1)
                self.output_layers[f"{name}_4"] = nn.Conv2d(nb_filter[0], config['channels'], kernel_size=1)
        else:
            self.output_layers = nn.ModuleDict()
            for name, config in self.output_heads.items():
                self.output_layers[name] = nn.Conv2d(nb_filter[0], config['channels'], kernel_size=1)

    def apply_activation(self, x, activation):
        if activation == 'sigmoid':
            return torch.sigmoid(x)
        elif activation == 'tanh':
            return torch.tanh(x)
        elif activation == 'relu':
            return torch.relu(x)
        return x

    def forward(self, input):
        x0_0 = self.conv0_0(input)
        x1_0 = self.conv1_0(self.pool(x0_0))
        x0_1 = self.conv0_1(torch.cat([x0_0, self.up(x1_0)], 1))

        x2_0 = self.conv2_0(self.pool(x1_0))
        x1_1 = self.conv1_1(torch.cat([x1_0, self.up(x2_0)], 1))
        x0_2 = self.conv0_2(torch.cat([x0_0, x0_1, self.up(x1_1)], 1))

        x3_0 = self.conv3_0(self.pool(x2_0))
        x2_1 = self.conv2_1(torch.cat([x2_0, self.up(x3_0)], 1))
        x1_2 = self.conv1_2(torch.cat([x1_0, x1_1, self.up(x2_1)], 1))
        x0_3 = self.conv0_3(torch.cat([x0_0, x0_1, x0_2, self.up(x1_2)], 1))

        x4_0 = self.conv4_0(self.pool(x3_0))
        x3_1 = self.conv3_1(torch.cat([x3_0, self.up(x4_0)], 1))
        x2_2 = self.conv2_2(torch.cat([x2_0, x2_1, self.up(x3_1)], 1))
        x1_3 = self.conv1_3(torch.cat([x1_0, x1_1, x1_2, self.up(x2_2)], 1))
        x0_4 = self.conv0_4(torch.cat([x0_0, x0_1, x0_2, x0_3, self.up(x1_3)], 1))

        outputs = {}
        if self.deep_supervision:
            for name, config in self.output_heads.items():
                if self.train_mode:
                    outputs[f"{name}_1"] = self.apply_activation(self.output_layers[f"{name}_1"](x0_1),
                                                                 config.get('activation'))
                    outputs[f"{name}_2"] = self.apply_activation(self.output_layers[f"{name}_2"](x0_2),
                                                                 config.get('activation'))
                    outputs[f"{name}_3"] = self.apply_activation(self.output_layers[f"{name}_3"](x0_3),
                                                                 config.get('activation'))
                    outputs[f"{name}_4"] = self.apply_activation(self.output_layers[f"{name}_4"](x0_4),
                                                                 config.get('activation'))
                    outputs[name] = outputs[f"{name}_4"]
                else:
                    outputs[name] = self.apply_activation(self.output_layers[f"{name}_4"](x0_4),
                                                          config.get('activation'))
        else:
            for name, config in self.output_heads.items():
                outputs[name] = self.apply_activation(self.output_layers[name](x0_4), config.get('activation'))

        return outputs


class MultiOutputNestedUNet_3Levels(nn.Module):
    def __init__(self, in_channels=1, output_heads: Dict[str, dict] = None, n_filter=32, deep_supervision=False,
                 dilation=False, train_mode=True, **kwargs):
        super().__init__()
        self.output_heads = output_heads or {'default': {'channels': 1, 'activation': 'sigmoid'}}
        self.deep_supervision = deep_supervision
        self.train_mode = train_mode
        self.dilation = dilation if dilation is not False else (1, 1, 1, 1)

        # Reduced from five to four entries to remove the fourth pooling level
        nb_filter = [n_filter, n_filter * 2, n_filter * 4, n_filter * 8]

        self.pool = nn.MaxPool2d(2, 2)
        self.up = nn.Upsample(scale_factor=2, mode='bilinear', align_corners=True)

        self.conv0_0 = VGGBlock(in_channels, nb_filter[0], nb_filter[0], self.dilation[0])
        self.conv1_0 = VGGBlock(nb_filter[0], nb_filter[1], nb_filter[1], self.dilation[1])
        self.conv2_0 = VGGBlock(nb_filter[1], nb_filter[2], nb_filter[2], self.dilation[2])
        self.conv3_0 = VGGBlock(nb_filter[2], nb_filter[3], nb_filter[3], self.dilation[3])

        self.conv0_1 = VGGBlock(nb_filter[0] + nb_filter[1], nb_filter[0], nb_filter[0])
        self.conv1_1 = VGGBlock(nb_filter[1] + nb_filter[2], nb_filter[1], nb_filter[1])
        self.conv2_1 = VGGBlock(nb_filter[2] + nb_filter[3], nb_filter[2], nb_filter[2])

        self.conv0_2 = VGGBlock(nb_filter[0] * 2 + nb_filter[1], nb_filter[0], nb_filter[0])
        self.conv1_2 = VGGBlock(nb_filter[1] * 2 + nb_filter[2], nb_filter[1], nb_filter[1])

        self.conv0_3 = VGGBlock(nb_filter[0] * 3 + nb_filter[1], nb_filter[0], nb_filter[0])

        if self.deep_supervision:
            self.output_layers = nn.ModuleDict()
            for name, config in self.output_heads.items():
                self.output_layers[f"{name}_1"] = nn.Conv2d(nb_filter[0], config['channels'], kernel_size=1)
                self.output_layers[f"{name}_2"] = nn.Conv2d(nb_filter[0], config['channels'], kernel_size=1)
                self.output_layers[f"{name}_3"] = nn.Conv2d(nb_filter[0], config['channels'], kernel_size=1)
        else:
            self.output_layers = nn.ModuleDict()
            for name, config in self.output_heads.items():
                self.output_layers[name] = nn.Conv2d(nb_filter[0], config['channels'], kernel_size=1)

    def apply_activation(self, x, activation):
        if activation == 'sigmoid':
            return torch.sigmoid(x)
        elif activation == 'tanh':
            return torch.tanh(x)
        elif activation == 'relu':
            return torch.relu(x)
        return x

    def forward(self, x):
        x0_0 = self.conv0_0(x)
        x1_0 = self.conv1_0(self.pool(x0_0))
        x0_1 = self.conv0_1(torch.cat([x0_0, self.up(x1_0)], 1))

        x2_0 = self.conv2_0(self.pool(x1_0))
        x1_1 = self.conv1_1(torch.cat([x1_0, self.up(x2_0)], 1))
        x0_2 = self.conv0_2(torch.cat([x0_0, x0_1, self.up(x1_1)], 1))

        x3_0 = self.conv3_0(self.pool(x2_0))
        x2_1 = self.conv2_1(torch.cat([x2_0, self.up(x3_0)], 1))
        x1_2 = self.conv1_2(torch.cat([x1_0, x1_1, self.up(x2_1)], 1))
        x0_3 = self.conv0_3(torch.cat([x0_0, x0_1, x0_2, self.up(x1_2)], 1))

        outputs = {}
        if self.deep_supervision:
            for name, config in self.output_heads.items():
                if self.train_mode:
                    outputs[f"{name}_1"] = self.apply_activation(self.output_layers[f"{name}_1"](x0_1),
                                                                 config.get('activation'))
                    outputs[f"{name}_2"] = self.apply_activation(self.output_layers[f"{name}_2"](x0_2),
                                                                 config.get('activation'))
                    outputs[f"{name}_3"] = self.apply_activation(self.output_layers[f"{name}_3"](x0_3),
                                                                 config.get('activation'))
                    outputs[name] = outputs[f"{name}_3"]
                else:
                    outputs[name] = self.apply_activation(self.output_layers[f"{name}_3"](x0_3),
                                                          config.get('activation'))
        else:
            for name, config in self.output_heads.items():
                outputs[name] = self.apply_activation(self.output_layers[name](x0_3), config.get('activation'))

        return outputs



import glob
import os
from typing import Union

import torch
import torch.optim as optim
from torch.utils.data import DataLoader, random_split
from tqdm import tqdm

from bio_image_unet.siam_unet import BCEDiceLoss
from . import logcoshTverskyLoss, TverskyLoss, weightedBCELoss
from .predict import Predict
from .siam_unet import Siam_UNet
from ..utils import get_device


class Trainer:
    """
    Class for training of neural network. Creates trainer object, training is started with .start().

    Parameters
    ----------
    dataset
        Training data, object of PyTorch Dataset class
    num_epochs : int
        Number of training epochs
    batch_size : int
        Batch size for training
    lr : float
        Learning rate
    n_filter : int
        Number of convolutional filters in first layer
    mode : str
        Mode to combine low-level features of T-1 and T ('max' or 'corr')
    val_split : float
        Validation split
    save_dir : str
        Path of directory to save trained networks
    save_name : str
        Base name for saving trained networks
    save_iter : bool
        If True, network state is save after each epoch
    load_weights : str, optional
        If not None, network state is loaded before training
    loss_function : str
        Loss function ('BCEDice', 'Tversky' or 'logcoshTversky')
    loss_params : Tuple[float, float]
        Parameter of loss function, depends on chosen loss function
    device : torch.device or str, optional
        Device to run the pytorch model on, defaults to 'auto', which selects CUDA or MPS if available.
    """
    def __init__(self, dataset, num_epochs, batch_size=4, lr=1e-3, n_filter=32, mode='max', val_split=0.2,
                 save_dir='./', save_name='model.pt', save_iter=False, loss_function='BCEDice',
                 loss_params=(1, 1), load_weights=None, device: Union[torch.device, str] = 'auto'):

        if device == 'auto':
            self.device = get_device()
        else:
            self.device = torch.device(device)

        self.model = Siam_UNet(n_filter=n_filter, mode=mode).to(self.device)

        self.data = dataset
        self.num_epochs = num_epochs
        self.batch_size = batch_size
        self.lr = lr
        self.n_filter = n_filter
        self.mode = mode
        self.best_loss = torch.tensor(float('inf'))
        self.save_iter = save_iter
        self.loss_function = loss_function
        self.loss_params = loss_params
        # split training and validation data
        num_val = int(len(dataset) * val_split)
        num_train = len(dataset) - num_val
        self.dim = dataset.dim_out
        train_data, val_data = random_split(dataset, [num_train, num_val])
        self.train_loader = DataLoader(train_data, batch_size=self.batch_size, pin_memory=True, drop_last=True)
        self.val_loader = DataLoader(val_data, batch_size=self.batch_size, pin_memory=True, drop_last=True)
        if loss_function == 'BCEDice':
            self.criterion = BCEDiceLoss(loss_params[0], loss_params[1])
        elif loss_function == 'Tversky':
            self.criterion = TverskyLoss(loss_params[0], loss_params[1])
        elif loss_function == 'logcoshTversky':
            self.criterion = logcoshTverskyLoss(loss_params[0], loss_params[1])
        elif loss_function == 'weightedBCELoss':
            self.criterion = weightedBCELoss(loss_params[0], loss_params[1])
        else:
            raise ValueError(f'Loss function "{loss_function}" not defined!')
        self.optimizer = torch.optim.Adam(self.model.parameters(), lr=self.lr)
        self.scheduler = optim.lr_scheduler.ReduceLROnPlateau(self.optimizer, 'min', patience=4, factor=0.1)
        self.save_dir = save_dir
        os.makedirs(self.save_dir, exist_ok=True)
        self.save_name = save_name

        if load_weights is not None:
            self.state = torch.load(load_weights)
            self.model.load_state_dict(self.state['state_dict'])

    def iterate(self, epoch, mode):
        if mode == 'train':
            print('\nStarting training epoch %s ...' % epoch)
            for i, batch_i in tqdm(enumerate(self.train_loader), total=len(self.train_loader), unit='batch'):
                x_i = batch_i['image'].view(self.batch_size, 1, self.dim[0], self.dim[1]).to(self.device)
                prev_x_i = batch_i['prev_image'].view(self.batch_size, 1, self.dim[0], self.dim[1]).to(self.device)
                y_i = batch_i['mask'].view(self.batch_size, 1, self.dim[0], self.dim[1]).to(self.device)
                # Forward pass: Compute predicted y by passing x to the model
                y_pred, y_logits = self.model(x_i, prev_x_i)
                # Compute loss
                loss = self.criterion(y_logits, y_i)
                # Zero gradients, perform a backward pass, and update the weights.
                self.optimizer.zero_grad()
                loss.backward()
                self.optimizer.step()

        elif mode == 'val':
            loss_list = []
            print('\nStarting validation epoch %s ...' % epoch)
            with torch.no_grad():
                for i, batch_i in tqdm(enumerate(self.val_loader), total=len(self.val_loader), unit='batch'):
                    x_i = batch_i['image'].view(self.batch_size, 1, self.dim[0], self.dim[1]).to(self.device)
                    prev_x_i = batch_i['prev_image'].view(self.batch_size, 1, self.dim[0], self.dim[1]).to(self.device)
                    y_i = batch_i['mask'].view(self.batch_size, 1, self.dim[0], self.dim[1]).to(self.device)
                    # Forward pass: Compute predicted y by passing x to the model
                    y_pred, y_logits = self.model(x_i, prev_x_i)
                    # Compute loss
                    loss = self.criterion(y_logits, y_i)
                    loss_list.append(loss.detach())
            val_loss = torch.stack(loss_list).mean()
            return val_loss

        torch.cuda.empty_cache()

    def start(self, test_data_path=None, result_path=None, test_resize_dim=(512, 512)):
        for epoch in range(self.num_epochs):
            self.iterate(epoch, 'train')
            self.state = {
                'epoch': epoch,
                'best_loss': self.best_loss,
                'state_dict': self.model.state_dict(),
                'optimizer': self.optimizer.state_dict(),
                'lr': self.lr,
                'loss': self.loss_function,
                'loss_params': self.loss_params,
                'n_filter': self.n_filter,
                'mode': self.mode,
                'augmentation': self.data.aug_factor,
                'clip_threshold': self.data.clip_threshold,
                'noise_amp': self.data.noise_amp,
                'brightness_contrast': self.data.brightness_contrast,
                'shiftscalerotate': self.data.shiftscalerotate,
            }
            with torch.no_grad():
                val_loss = self.iterate(epoch, 'val')
                self.scheduler.step(val_loss)
            if val_loss < self.best_loss:
                print(f'\nEpoch {epoch}: Validation loss improved from {round(self.best_loss.item(), 5)} to '
                      f'{round(val_loss.item(), 5)} - saving model state')
                self.state['best_loss'] = self.best_loss = val_loss
                torch.save(self.state, self.save_dir + '/' + self.save_name)
            else:
                print(f'\nEpoch {epoch}: Validation loss did not improve from {round(self.best_loss.item(), 5)}')
            if self.save_iter:
                torch.save(self.state, self.save_dir + '/' + f'model_epoch_{epoch}.pt')

            if test_data_path is not None:
                os.makedirs(result_path, exist_ok=True)
                files = glob.glob(test_data_path + '*.tif')
                for i, file in enumerate(files):
                    Predict(file, result_path + os.path.basename(file) + f'epoch_{epoch}.tif',
                            self.save_dir + '/' + f'model_epoch_{epoch}.pt', resize_dim=test_resize_dim,
                            invert=False, show_progress=False)



import os.path
from typing import Union

import numpy as np
import tifffile
import torch

from ..progress import ProgressNotifier
from .multi_output_nested_unet import MultiOutputNestedUNet, MultiOutputNestedUNet_3Levels
from ..utils import get_device


class Predict:
    """Class for prediction of movies and images using a U-Net model."""

    def __init__(self, imgs, model_params, result_path=None, network=MultiOutputNestedUNet, max_patch_size=(1024, 1024),
                 batch_size=1, normalization_mode='single', clip_threshold=(0., 99.98), add_tile=0,
                 compress_tif=False, show_progress=True, device: Union[torch.device, str] = 'auto',
                 progress_notifier: ProgressNotifier = ProgressNotifier.progress_notifier_tqdm()):
        """
        Initialize the Predict class for performing predictions on TIFF files using a standard 2D U-Net.

        The process involves:
        1) Loading and preprocessing the input images (including normalization).
        2) Resizing images into patches based on the specified dimensions.
        3) Performing predictions using the U-Net model.
        4) Stitching predicted patches together and averaging overlapping regions.

        Parameters
        ----------
        imgs : ndarray or str
            Images to predict. If a string is provided, it attempts to load images from a TIFF file.
        model_params : str
            Path to the U-Net model parameters (.pt file).
        result_path : str, optional
            Path to save the prediction results as a TIFF file. If None, results are stored in the 'result' attribute.
        network : Network, optional
            Network architecture. Default is MultiOutputUnet
        max_patch_size : Tuple[int, int]
            Maximal dimensions to resize images for prediction.
        batch_size : int
            Number of images to process in each batch during prediction.
        normalization_mode : str
            Mode for intensity normalization for 3D stacks prior to prediction. Options are:
            - 'single': Normalize each image individually.
            - 'all': Normalize based on the histogram of the entire stack.
            - 'first': Normalize based on the histogram of the first image in the stack.
        clip_threshold : Tuple[float, float]
            Thresholds for clipping image intensity before prediction.
        add_tile : int, optional
            Additional tiles to add for splitting large images to increase overlap.
        compress_tif : bool, optional
            Whether to compress tif files to save storage. Default is False.
        show_progress : bool
            Whether to display a progress bar during prediction.
        device : torch.device or str, optional
            Device to run the PyTorch model on. Defaults to 'auto', which selects CUDA or MPS if available.
        progress_notifier : ProgressNotifier, optional
            Wrapper to display a progress notifier in the GUI using tqdm.
        """
        if device == 'auto':
            self.device = get_device()
        else:
            self.device = torch.device(device)

        self.precision = torch.float16 if self.device.type == 'cuda' or self.device.type == 'mps' else torch.float32

        if isinstance(imgs, str):
            imgs = tifffile.imread(imgs)

        self.max_patch_size = max_patch_size
        self.batch_size = batch_size
        self.add_tile = add_tile
        self.normalization_mode = normalization_mode
        self.clip_threshold = clip_threshold
        self.result_path = result_path
        self.compress_tif = compress_tif
        self.show_progress = show_progress

        # read, preprocess and split data
        imgs = imgs.astype('float32')
        imgs = self.__reshape_data(imgs)
        imgs = self.__preprocess(imgs)
        patches = self.__split(imgs)
        del imgs

        # load model and predict data
        self.model_params = torch.load(model_params, map_location=self.device)
        self.model = network(in_channels=self.model_params['in_channels'],
                             n_filter=self.model_params['n_filter'],
                             output_heads=self.model_params['output_heads'],
                             deep_supervision=self.model_params.get('deep_supervision', False),
                             train_mode=False).to(self.device)
        self.model.load_state_dict(self.model_params['state_dict'])
        self.model.to(self.precision)
        self.model.eval()
        self.target_keys = self.model_params['output_heads'].keys()
        result_patches = self.__predict(patches, self.batch_size, progress_notifier)
        del patches, self.model

        # stitch patches (mean of overlapped regions)
        result = self.__stitch(result_patches)
        del result_patches, self.model_params

        if self.result_path is not None:
            # save as .tif files
            for target_key in self.target_keys:
                target_file = self.result_path + target_key + '.tif' if os.path.exists(self.result_path) else (
                        self.result_path + '_' + target_key + '.tif')
                if self.compress_tif:
                    tifffile.imwrite(target_file, result[target_key], compression='deflate',
                                     compressionargs={'level': 6}, predictor=3)
                else:
                    tifffile.imwrite(target_file, result[target_key],)
            del result
            self.result = None
        else:
            self.result = result
        if torch.cuda.is_available():
            torch.cuda.empty_cache()

    def __reshape_data(self, imgs):
        self.imgs_shape = imgs.shape
        if len(self.imgs_shape) == 2:  # if single image
            imgs = np.expand_dims(imgs, axis=0)
            self.imgs_shape = imgs.shape
        return imgs

    def __preprocess(self, imgs):
        if self.normalization_mode == 'single':
            for i, img in enumerate(imgs):
                img = np.clip(img, a_min=np.nanpercentile(img, self.clip_threshold[0]),
                              a_max=np.percentile(img, self.clip_threshold[1]))
                img = img - np.min(img)
                img = img / np.max(img)
                imgs[i] = img
        elif self.normalization_mode == 'first':
            clip_threshold = (np.nanpercentile(imgs[0], self.clip_threshold[0]),
                              np.percentile(imgs[0], self.clip_threshold[1]))
            imgs = np.clip(imgs, clip_threshold[0], clip_threshold[1])
            imgs = imgs - np.min(imgs)
            imgs = imgs / np.max(imgs)
        elif self.normalization_mode == 'all':
            clip_threshold = (np.nanpercentile(imgs, self.clip_threshold[0]),
                              np.percentile(imgs, self.clip_threshold[1]))
            imgs = np.clip(imgs, clip_threshold[0], clip_threshold[1])
            imgs = imgs - np.min(imgs)
            imgs = imgs / np.max(imgs)
        else:
            raise ValueError(f'normalization_mode {self.normalization_mode} not valid!')
        return imgs

    def __split(self, imgs):
        # Calculate the actual patch size
        patch_height = min(self.imgs_shape[1], self.max_patch_size[0])
        patch_width = min(self.imgs_shape[2], self.max_patch_size[1])

        # Adjust patch size to be divisible by 16
        patch_height = ((patch_height + 15) // 16) * 16
        patch_width = ((patch_width + 15) // 16) * 16

        self.patch_size = (patch_height, patch_width)

        # Calculate number of patches
        self.N_x = int(np.ceil(self.imgs_shape[1] / patch_height)) + self.add_tile
        self.N_y = int(np.ceil(self.imgs_shape[2] / patch_width)) + self.add_tile
        self.N_per_img = self.N_x * self.N_y
        self.N = self.N_per_img * self.imgs_shape[0]

        # Pad images if needed
        pad_x = max(patch_height - self.imgs_shape[1], 0)
        pad_y = max(patch_width - self.imgs_shape[2], 0)
        imgs = np.pad(imgs, ((0, 0), (0, pad_x), (0, pad_y)), 'reflect')

        # Calculate starting indices (preserved for stitching)
        self.X_start = np.linspace(0, imgs.shape[1] - patch_height, self.N_x).astype('uint16')
        self.Y_start = np.linspace(0, imgs.shape[2] - patch_width, self.N_y).astype('uint16')

        # Generate patches using vectorized operations
        patches = np.lib.stride_tricks.sliding_window_view(imgs, self.patch_size, axis=(1, 2))
        patches = patches[:, ::self.X_start[1] if self.N_x > 1 else 1, ::self.Y_start[1] if self.N_y > 1 else 1]
        patches = patches.reshape(-1, *self.patch_size)

        return patches

    def __predict(self, patches, batch_size=1, progress_notifier=None):
        with torch.no_grad():
            # Initialize result_patches with correct dimensions
            result_patches = {
                target_key: np.zeros(
                    (patches.shape[0], self.model_params['output_heads'][target_key]['channels'], *self.patch_size),
                    dtype='float16'
                )
                for target_key in self.target_keys
            }

            num_batches = int(np.ceil(patches.shape[0] / batch_size))
            _progress_notifier = range(num_batches)
            if self.show_progress and progress_notifier:
                _progress_notifier = progress_notifier.iterator(_progress_notifier)

            for batch_idx in _progress_notifier:
                start_idx = batch_idx * batch_size
                end_idx = min(start_idx + batch_size, patches.shape[0])

                # Prepare batch patches
                batch_patches = torch.tensor(patches[start_idx:end_idx], dtype=self.precision).to(self.device)
                batch_patches = batch_patches.view(
                    -1, self.model_params['in_channels'], self.patch_size[0], self.patch_size[1]
                )

                # Get model predictions
                batch_results = self.model(batch_patches)

                # Assign predictions to result_patches
                for target_key in self.target_keys:
                    # Extract model output for this target key
                    batch_results_target = batch_results[target_key].cpu().numpy()

                    # Ensure shapes match before assignment
                    expected_shape = result_patches[target_key][start_idx:end_idx].shape
                    if batch_results_target.shape != expected_shape:
                        raise RuntimeError(
                            f"Shape mismatch for target key '{target_key}': "
                            f"predicted {batch_results_target.shape}, expected {expected_shape}"
                        )

                    result_patches[target_key][start_idx:end_idx] = batch_results_target
                del batch_patches, batch_results
                torch.cuda.empty_cache()

        return result_patches

    def __stitch(self, result_patches, safe_margin=20):
        result = {}

        for target_key in self.target_keys:
            result_patches_target = result_patches[target_key]
            result_target = np.zeros(
                shape=(self.imgs_shape[0], self.model_params['output_heads'][target_key]['channels'],
                       max(self.patch_size[0], self.imgs_shape[1]),
                       max(self.patch_size[1], self.imgs_shape[2])),
                dtype='float32'
            )
            weight = np.zeros_like(result_target)

            for i in range(self.imgs_shape[0]):
                stack_result_i = result_patches_target[
                                 i * self.N_per_img:(i + 1) * self.N_per_img
                                 ].reshape(self.N_x, self.N_y, *result_patches_target.shape[1:])

                for j, x_start in enumerate(self.X_start):
                    for k, y_start in enumerate(self.Y_start):
                        x_slice = slice(x_start, x_start + self.patch_size[0])
                        y_slice = slice(y_start, y_start + self.patch_size[1])
                        patch = stack_result_i[j, k]

                        # Create weight mask for this patch
                        patch_weight = np.ones_like(patch)

                        # Only apply margin where overlaps exist
                        if j > 0:  # Left margin
                            patch_weight[..., :safe_margin, :] = 0
                        if j < self.N_x - 1:  # Right margin
                            patch_weight[..., -safe_margin:, :] = 0
                        if k > 0:  # Top margin
                            patch_weight[..., :safe_margin] = 0
                        if k < self.N_y - 1:  # Bottom margin
                            patch_weight[..., -safe_margin:] = 0

                        # Add weighted contribution
                        result_target[i, :, x_slice, y_slice] += patch * patch_weight
                        weight[i, :, x_slice, y_slice] += patch_weight

            # Normalize using the accumulated weights
            np.divide(result_target, weight, out=result_target, where=weight > 0)

            # Handle completely unweighted areas (should only happen at very edges)
            result_target[weight == 0] = result_patches_target.mean()

            # Crop to original image size
            result_target = result_target[:, :, :self.imgs_shape[1], :self.imgs_shape[2]]
            result[target_key] = np.squeeze(result_target)

        return result



import torch
import torch.nn as nn
import torch.nn.functional as F


class BCELoss(nn.Module):
    """
    Binary Cross-Entropy Loss.

    Parameters
    ----------
    weight : torch.Tensor, optional
        A manual rescaling weight given to the loss of each batch element.
    size_average : bool, optional
        By default, the losses are averaged over each loss element in the batch.
        If size_average is set to False, the losses are instead summed for each minibatch (default is True).
    """

    def __init__(self, weight=None, size_average=True):
        super(BCELoss, self).__init__()
        self.bce_loss = nn.BCEWithLogitsLoss(weight=weight, reduction='mean' if size_average else 'sum')

    def forward(self, logits, targets):
        """
        Forward pass.

        Parameters
        ----------
        logits : torch.Tensor
            The predicted outputs.
        targets : torch.Tensor
            The ground truth labels.

        Returns
        -------
        torch.Tensor
            The binary cross-entropy loss value.
        """
        return self.bce_loss(logits, targets)


class SoftDiceLoss(nn.Module):
    """
    Soft Dice Loss.

    Parameters
    ----------
    smooth : float, optional
        Smoothing factor to avoid division by zero (default is 1.0).
    """

    def __init__(self, smooth=1.0):
        super(SoftDiceLoss, self).__init__()
        self.smooth = smooth

    def forward(self, logits, targets):
        """
        Forward pass.

        Parameters
        ----------
        logits : torch.Tensor
            The predicted outputs.
        targets : torch.Tensor
            The ground truth labels.

        Returns
        -------
        torch.Tensor
            The soft dice loss value.
        """
        probs = torch.sigmoid(logits)
        num = targets.size(0)
        m1 = probs.view(num, -1)
        m2 = targets.view(num, -1)
        intersection = (m1 * m2).sum(1)
        score = 2. * (intersection + self.smooth) / (m1.sum(1) + m2.sum(1) + self.smooth)
        return 1 - score.mean()


class BCEDiceLoss(nn.Module):
    """
    Combines Binary Cross-Entropy Loss and Dice Loss.

    Parameters
    ----------
    alpha : float
        Weight for the Binary Cross-Entropy Loss component.
    beta : float
        Weight for the Dice Loss component.
    """

    def __init__(self, alpha, beta):
        super(BCEDiceLoss, self).__init__()
        self.bce = BCELoss()
        self.dice = SoftDiceLoss()
        self.alpha = alpha
        self.beta = beta

    def forward(self, logits, targets):
        """
        Forward pass.

        Parameters
        ----------
        logits : torch.Tensor
            The predicted outputs.
        targets : torch.Tensor
            The ground truth labels.

        Returns
        -------
        torch.Tensor
            The combined loss value.
        """
        return self.alpha * self.bce(logits, targets) + self.beta * self.dice(logits, targets)


class logcoshDiceLoss(nn.Module):
    """
    Log-Cosh Dice Loss.

    Combines the Log-Cosh function with Dice Loss for improved stability.
    """

    def __init__(self):
        super(logcoshDiceLoss, self).__init__()
        self.dice = SoftDiceLoss()

    def forward(self, logits, targets):
        """
        Forward pass.

        Parameters
        ----------
        logits : torch.Tensor
            The predicted outputs.
        targets : torch.Tensor
            The ground truth labels.

        Returns
        -------
        torch.Tensor
            The Log-Cosh Dice loss value.
        """
        x = self.dice(logits, targets)
        return torch.log((torch.exp(x) + torch.exp(-x)) / 2)


class TverskyLoss(nn.Module):
    """
    Tversky Loss.

    Parameters
    ----------
    alpha : float, optional
        Weight of false positives (default is 0.5).
    beta : float, optional
        Weight of false negatives (default is 0.5).
    smooth : float, optional
        Smoothing factor to avoid division by zero (default is 1).
    """

    def __init__(self, alpha=0.5, beta=0.5, smooth=1):
        super(TverskyLoss, self).__init__()
        self.alpha = alpha
        self.beta = beta
        self.smooth = smooth

    def forward(self, inputs, targets):
        """
        Forward pass.

        Parameters
        ----------
        inputs : torch.Tensor
            The predicted outputs.
        targets : torch.Tensor
            The ground truth labels.

        Returns
        -------
        torch.Tensor
            The Tversky loss value.
        """
        inputs = torch.sigmoid(inputs)
        inputs = inputs.view(-1)
        targets = targets.view(-1)

        # True Positives, False Positives & False Negatives
        TP = (inputs * targets).sum()
        FP = ((1 - targets) * inputs).sum()
        FN = (targets * (1 - inputs)).sum()

        Tversky = (TP + self.smooth) / (TP + self.alpha * FP + self.beta * FN + self.smooth)

        return 1 - Tversky


class logcoshTverskyLoss(nn.Module):
    """
    Log-Cosh Tversky Loss.

    Parameters
    ----------
    alpha : float, optional
        Weight of false positives (default is 0.5).
    beta : float, optional
        Weight of false negatives (default is 0.5).
    smooth : float, optional
        Smoothing factor to avoid division by zero (default is 1).
    """

    def __init__(self, alpha=0.5, beta=0.5, smooth=1):
        super(logcoshTverskyLoss, self).__init__()
        self.alpha = alpha
        self.beta = beta
        self.smooth = smooth

    def forward(self, inputs, targets):
        """
        Forward pass.

        Parameters
        ----------
        inputs : torch.Tensor
            The predicted outputs.
        targets : torch.Tensor
            The ground truth labels.

        Returns
        -------
        torch.Tensor
            The log-cosh Tversky loss value.
        """
        inputs = torch.sigmoid(inputs)
        inputs = inputs.view(-1)
        targets = targets.view(-1)

        # True Positives, False Positives & False Negatives
        TP = (inputs * targets).sum()
        FP = ((1 - targets) * inputs).sum()
        FN = (targets * (1 - inputs)).sum()

        Tversky = (TP + self.smooth) / (TP + self.alpha * FP + self.beta * FN + self.smooth)

        return torch.log(torch.cosh(1 - Tversky))


class TemporalConsistencyLoss(nn.Module):
    def __init__(self):
        super(TemporalConsistencyLoss, self).__init__()
        self.loss_fn = nn.L1Loss()

    def forward(self, predictions):
        """
        predictions: Tensor of shape (B, C, Z, X, Y)
                     where Z is the temporal dimension
        """
        # Compute loss between consecutive frames along the Z dimension
        loss = self.loss_fn(predictions[:, :, 1:, :, :], predictions[:, :, :-1, :, :])

        return loss


class BCEDiceTemporalLoss(nn.Module):
    def __init__(self, loss_params=(1.0, 0.1)):
        """
        Combined loss for 3D/temporal segmentation with BCEDice and temporal consistency.

        Args:
            loss_params (tuple): Weights for (BCEDice, TemporalConsistency) losses.
                                Default is (1.0, 0.1).
        """
        super(BCEDiceTemporalLoss, self).__init__()
        self.bce_dice_loss = BCEDiceLoss(1, 1)  # Your existing implementation
        self.temporal_consistency_loss = TemporalConsistencyLoss()
        self.loss_params = loss_params

    def forward(self, predictions, targets):
        """
        Args:
            predictions: Tensor of shape (batch_size, frames, channels, height, width)
            targets: Tensor of shape (batch_size, frames, channels, height, width)
        """
        # Compute BCEDice loss using your implementation
        bce_dice_loss_value = self.bce_dice_loss(predictions, targets)

        # Compute Temporal Consistency Loss
        temporal_consistency_loss_value = self.temporal_consistency_loss(predictions)

        # Combine losses with weights from the tuple
        total_loss = (
                self.loss_params[0] * bce_dice_loss_value +
                self.loss_params[1] * temporal_consistency_loss_value
        )

        return total_loss



import torch
import torch.nn as nn
import torch.nn.functional as F


# Classification loss functions

class BCEDiceLoss(nn.Module):
    def __init__(self, bce_weight=0.5, dice_weight=0.5):
        super(BCEDiceLoss, self).__init__()
        self.bce_weight = bce_weight
        self.dice_weight = dice_weight
        self.bce = nn.BCELoss()  # Use BCELoss for probabilities

    def forward(self, inputs, targets):
        assert torch.all((inputs >= 0) & (inputs <= 1)), "Inputs must be between 0 and 1"
        assert torch.all((targets >= 0) & (targets <= 1)), "Targets must be between 0 and 1"

        # Compute BCE loss with probabilities
        bce_loss = self.bce(inputs, targets)

        # Dice loss calculation
        smooth = 1e-5
        intersection = (inputs * targets).sum()
        dice_loss = 1 - (2. * intersection + smooth) / (inputs.sum() + targets.sum() + smooth)

        # Combine BCE and Dice losses
        return self.bce_weight * bce_loss + self.dice_weight * dice_loss


class TverskyLoss(nn.Module):
    def __init__(self, alpha=0.5, beta=0.5, smooth=1):
        super(TverskyLoss, self).__init__()
        self.alpha = alpha
        self.beta = beta
        self.smooth = smooth

    def forward(self, inputs, targets):
        inputs = inputs.view(-1)
        targets = targets.view(-1)

        # True Positives, False Positives & False Negatives
        TP = (inputs * targets).sum()
        FP = ((1 - targets) * inputs).sum()
        FN = (targets * (1 - inputs)).sum()

        Tversky = (TP + self.smooth) / (TP + self.alpha * FP + self.beta * FN + self.smooth)

        return 1 - Tversky


class logcoshTverskyLoss(nn.Module):
    def __init__(self, alpha=0.5, beta=0.5, smooth=1):
        super(logcoshTverskyLoss, self).__init__()
        self.alpha = alpha
        self.beta = beta
        self.smooth = smooth

    def forward(self, inputs, targets):
        inputs = inputs.view(-1)
        targets = targets.view(-1)

        # True Positives, False Positives & False Negatives
        TP = (inputs * targets).sum()
        FP = ((1 - targets) * inputs).sum()
        FN = (targets * (1 - inputs)).sum()

        Tversky = (TP + self.smooth) / (TP + self.alpha * FP + self.beta * FN + self.smooth)

        return torch.log(torch.cosh(1 - Tversky))


# Regression loss functions

class MSELoss(nn.Module):
    def __init__(self):
        super(MSELoss, self).__init__()

    def forward(self, inputs, targets):
        return ((inputs - targets) ** 2).mean()


class MAELoss(nn.Module):
    def __init__(self):
        super(MAELoss, self).__init__()

    def forward(self, inputs, targets):
        return torch.abs(inputs - targets).mean()


class HuberLoss(nn.Module):
    def __init__(self, delta=1.0):
        super(HuberLoss, self).__init__()
        self.delta = delta

    def forward(self, inputs, targets):
        diff = torch.abs(inputs - targets)
        loss = torch.where(diff < self.delta, 0.5 * diff ** 2, self.delta * (diff - 0.5 * self.delta))
        return loss.mean()


def gradient_loss(pred, target):
    """Calculate gradient loss comparing spatial derivatives of pred and target."""
    # Calculate gradients in y and x directions
    dy_true, dx_true = torch.gradient(target, dim=(-2, -1))  # dim=-2 is y direction, dim=-1 is x direction
    dy_pred, dx_pred = torch.gradient(pred, dim=(-2, -1))

    # Calculate MSE between gradients
    gradient_loss_y = F.mse_loss(dy_pred, dy_true)
    gradient_loss_x = F.mse_loss(dx_pred, dx_true)

    return gradient_loss_y + gradient_loss_x


class DistanceGradientLoss(torch.nn.Module):
    """Combined loss for distance regression with gradient preservation."""

    def __init__(self, alpha=1):
        super().__init__()
        self.alpha = alpha

    def forward(self, pred, target):
        # Distance loss (MSE)
        distance_loss = F.mse_loss(pred, target)

        # Gradient loss
        grad_loss = gradient_loss(pred, target)

        # Combined loss
        total_loss = distance_loss + self.alpha * grad_loss

        return total_loss


class WeightedDistanceGradientLoss(torch.nn.Module):
    def __init__(self, alpha=1.0, beta=0.5):
        super().__init__()
        self.alpha = alpha  # gradient loss weight
        self.beta = beta  # weight for non-zero regions

    def forward(self, pred, target):
        # Create weight mask for non-zero regions
        weights = torch.where(target > 0, self.beta, 1.0 - self.beta)

        # Weighted distance loss (combining MSE and MAE for robustness)
        mse_loss = F.mse_loss(pred * weights, target * weights, reduction='mean')
        mae_loss = F.l1_loss(pred * weights, target * weights, reduction='mean')
        distance_loss = mse_loss + mae_loss

        # Weighted gradient loss
        grad_loss = gradient_loss(pred * weights, target * weights)

        return distance_loss + self.alpha * grad_loss


class WeightedVectorFieldLoss(nn.Module):
    def __init__(self, beta=0.5, magnitude_weight=0.3):
        super().__init__()
        self.beta = beta
        self.magnitude_weight = magnitude_weight

    def forward(self, pred_vectors, true_vectors):
        """
        Args:
            pred_vectors: Predicted vector field (B, 2, H, W)
            true_vectors: Ground truth vector field (B, 2, H, W)
        """
        # Create mask where vectors are valid (both components non-zero)
        mask = ~((true_vectors[:, 0] == 0) & (true_vectors[:, 1] == 0))

        # Create weights
        weights = torch.where(mask, self.beta, 1.0 - self.beta)

        # Component-wise losses
        mse_loss = F.mse_loss(pred_vectors * weights[:, None],
                              true_vectors * weights[:, None],
                              reduction='mean')
        mae_loss = F.l1_loss(pred_vectors * weights[:, None],
                             true_vectors * weights[:, None],
                             reduction='mean')

        # Magnitude loss
        pred_magnitude = torch.sum(pred_vectors ** 2, dim=1)
        true_magnitude = torch.sum(true_vectors ** 2, dim=1)
        magnitude_loss = F.mse_loss(pred_magnitude * weights,
                                    true_magnitude * weights,
                                    reduction='mean')

        return mse_loss + mae_loss + self.magnitude_weight * magnitude_loss



import glob
import os
import shutil
from skimage import morphology
from albumentations import (
    ShiftScaleRotate, GaussNoise,
    RandomBrightnessContrast, Compose, RandomRotate90)

import numpy as np
import tifffile
import torch
from torch.utils.data import Dataset


class DataProcess(Dataset):
    """
    Class for training data generation (processing, splitting, augmentation)

    Create training data object for network training

    1) Create folder structure for training data
    2) Move and preprocess training volumes
    3) Split input volumes into patches
    4) Augment training data
    5) Create object of PyTorch Dataset class for training

    Parameters
    ----------
    source_dir : Tuple[str, str]
        Path of training data [volumes, labels]. volumes need to be tif files.
    dim_out : Tuple[int, int]
        Resize dimensions of volumes for training
    aug_factor : int
        Factor of volume augmentation
    data_path : str
        Base path of temporary directories for training data
    dilate_mask
        Radius of binary dilation of masks [-2, -1, 0, 1, 2]
    dilate_kernel : str
        Dilation kernel ('disk' or 'square')
    add_patch : int
        Add additional patch for splitting volumes into patches with more overlapping patches
    val_split : float
        Validation split for training
    invert : bool
        If True, greyscale binary labels is inverted
    skeletonize : bool
        If True, binary labels are skeletonized
    clip_threshold : Tuple[float, float]
        Clip thresholds for intensity normalization of volumes
    shiftscalerotate : [float, float, float]
        Shift, scale and rotate during augmentation
    noise_amp : float
        Amplitude of Gaussian noise for augmentation
    brightness_contrast : Tuple[float, float]
        Adapt brightness and contrast of volumes during augmentation
    rescale : float, optional
        Rescale all volumes and labels by factor rescale
    create : bool, optional
        If False, existing data set in data_path is used
    """

    def __init__(self, source_dir, dim_out=(128, 128, 128), aug_factor=10, data_path='../data/', dilate_mask=0, dilate_kernel='disk', add_patch=0,
                 val_split=0.2, invert=False, skeletonize=False, clip_threshold=(0.2, 99.8), shiftscalerotate=(0, 0, 0),
                 noise_amp=10, brightness_contrast=(0.25, 0.25), create=True):

        self.source_dir = source_dir
        self.create = create
        self.data_path = data_path
        self.dim_out = dim_out
        self.skeletonize = skeletonize
        self.invert = invert
        self.clip_threshold = clip_threshold
        self.add_patch = add_patch
        self.aug_factor = aug_factor
        self.shiftscalerotate = shiftscalerotate
        self.brightness_contrast = brightness_contrast
        self.noise_amp = noise_amp
        self.dilate_mask = dilate_mask
        self.dilate_kernel = dilate_kernel
        self.val_split = val_split
        self.mode = 'train'

        self.__make_dirs()
        if create:
            self.__move_and_edit()
            self.__merge_volumes()
            self.__split()
            if self.aug_factor is not None:
                self.__augment()

    def __make_dirs(self):
        self.volume_path = self.data_path + '/volume/'
        self.mask_path = self.data_path + '/mask/'
        self.merge_path = self.data_path + '/merge/'
        self.split_merge_path = self.data_path + '/split/merge/'
        self.split_volume_path = self.data_path + '/split/volume/'
        self.split_mask_path = self.data_path + '/split/mask/'
        self.aug_volume_path = self.data_path + '/augmentation/aug_volume/'
        self.aug_mask_path = self.data_path + '/augmentation/aug_mask/'

        # delete old files
        if self.create and os.path.exists(self.data_path):
            shutil.rmtree(self.data_path)

        # make folders
        os.makedirs(self.volume_path, exist_ok=True)
        os.makedirs(self.mask_path, exist_ok=True)
        os.makedirs(self.merge_path, exist_ok=True)
        os.makedirs(self.split_merge_path, exist_ok=True)
        os.makedirs(self.split_volume_path, exist_ok=True)
        os.makedirs(self.split_mask_path, exist_ok=True)
        os.makedirs(self.aug_volume_path, exist_ok=True)
        os.makedirs(self.aug_mask_path, exist_ok=True)

    def __move_and_edit(self):
        # create volume data
        files_volume = glob.glob(self.source_dir[0] + '*')
        for file_i in files_volume:
            vol_i = tifffile.imread(file_i).astype(np.float32)
            # clip and normalize (0,255)
            vol_i = np.clip(vol_i, a_min=np.percentile(vol_i, self.clip_threshold[0]),
                            a_max=np.percentile(vol_i, self.clip_threshold[1]))
            vol_i = (vol_i - np.nanmin(vol_i)) / (np.nanmax(vol_i) - np.nanmin(vol_i)) * 255
            vol_i = vol_i.astype('uint8')
            save_i = os.path.splitext(os.path.basename(file_i))[0]
            save_i = save_i.replace(' ', '_')
            tifffile.imwrite(self.volume_path + save_i + '.tif', vol_i)
        # create masks
        files_mask = glob.glob(self.source_dir[1] + '*')
        print('%s files found' % len(files_mask))
        for file_i in files_mask:
            mask_i = tifffile.imread(file_i)
            for j, ch_j in enumerate(mask_i):  # iterate frames / slices
                if self.skeletonize:
                    ch_j = ch_j > 1
                    ch_j = morphology.skeletonize(ch_j) * 255
                if self.dilate_kernel == 'disk':
                    kernel = morphology.disk
                elif self.dilate_kernel == 'square':
                    kernel = morphology.square
                else:
                    raise ValueError(f'Dilate kernel {self.dilate_kernel} unknown!')
                if self.dilate_mask > 0 and self.dilate_mask:
                    ch_j = morphology.erosion(ch_j, kernel(self.dilate_mask))
                elif self.dilate_mask < 0:
                    ch_j = morphology.dilation(ch_j, kernel(-self.dilate_mask))
                if self.invert:
                    ch_j = 255 - ch_j
                mask_i[j] = ch_j
            mask_i = mask_i.astype('uint8')
            save_i = os.path.splitext(os.path.basename(file_i))[0]
            save_i = save_i.replace(' ', '_')
            tifffile.imwrite(self.mask_path + save_i + '.tif', mask_i.astype('int8'))

    def __merge_volumes(self):
        self.mask_files = glob.glob(self.data_path + '/mask/*.tif')
        self.volume_files = glob.glob(self.data_path + '/volume/*.tif')

        if len(self.mask_files) != len(self.volume_files):
            raise ValueError('Number of ground truth does not match number of volume stacks')

        for i, file_i in enumerate(self.mask_files):
            basename_i = os.path.basename(file_i)
            mask_i = tifffile.imread(self.data_path + '/mask/' + basename_i)
            vol_i = tifffile.imread(self.data_path + '/volume/' + basename_i)
            merge = np.zeros((mask_i.shape[0], mask_i.shape[1], mask_i.shape[2], 2), dtype='uint8')
            merge[:, :, :, 0] = vol_i
            merge[:, :, :, 1] = mask_i
            tifffile.imwrite(self.merge_path + str(i) + '.tif', merge)

    def __split(self):
        self.merges = glob.glob(self.merge_path + '*.tif')
        n = 0
        for i in range(len(self.merges)):
            merge = tifffile.imread(self.merge_path + str(i) + '.tif')
            dim_in = merge.shape
            # padding if dim_in < dim_out
            z_gap = max(0, self.dim_out[0] - dim_in[0])
            x_gap = max(0, self.dim_out[1] - dim_in[1])
            y_gap = max(0, self.dim_out[2] - dim_in[2])
            merge = np.pad(merge, ((0, z_gap), (0, x_gap), (0, y_gap), (0, 0)), 'reflect')
            # number of patches in x and y
            dim_in = merge.shape
            N_z = int(np.ceil(dim_in[0] / self.dim_out[0]))
            N_x = int(np.ceil(dim_in[1] / self.dim_out[1]))
            N_y = int(np.ceil(dim_in[2] / self.dim_out[2]))
            N_x += self.add_patch if N_z > 1 else 0
            N_x += self.add_patch if N_x > 1 else 0
            N_y += self.add_patch if N_y > 1 else 0
            # starting indices of patches
            Z_start = np.linspace(0, dim_in[0] - self.dim_out[0], N_z).astype('int16')
            X_start = np.linspace(0, dim_in[1] - self.dim_out[1], N_x).astype('int16')
            Y_start = np.linspace(0, dim_in[2] - self.dim_out[2], N_y).astype('int16')
            for j in range(N_z):
                for k in range(N_x):
                    for l in range(N_y):
                        patch_ij = merge[Z_start[j]:Z_start[j] + self.dim_out[0],
                                   X_start[k]: X_start[k] + self.dim_out[1],
                                   Y_start[l]: Y_start[l] + self.dim_out[2], :]
                        volume_ij = patch_ij[:, :, :, 0]
                        mask_ij = patch_ij[:, :, :, 1]

                        tifffile.imwrite(self.split_merge_path + f'{n}.tif', patch_ij)
                        tifffile.imwrite(self.split_mask_path + f'{n}.tif', mask_ij)
                        tifffile.imwrite(self.split_volume_path + f'{n}.tif', volume_ij)
                        n += 1

    def __augment(self, p=0.8):
        aug_pipeline = Compose(transforms=[
            RandomRotate90(p=1.0),
            ShiftScaleRotate(self.shiftscalerotate[0], self.shiftscalerotate[1], self.shiftscalerotate[2]),
            GaussNoise(var_limit=(self.noise_amp, self.noise_amp), p=0.3),
            RandomBrightnessContrast(brightness_limit=self.brightness_contrast[0],
                                     contrast_limit=self.brightness_contrast[1], p=0.5),
        ],
            p=p)

        patches_volume = glob.glob(self.split_volume_path + '*.tif')
        patches_mask = glob.glob(self.split_mask_path + '*.tif')

        n_patches = len(patches_volume)
        k = 0
        for i in range(n_patches):
            volume_i = tifffile.imread(patches_volume[i])
            mask_i = tifffile.imread(patches_mask[i])
            volume_i = volume_i.transpose(1, 2, 0)
            mask_i = mask_i.transpose(1, 2, 0)

            data_i = {'image': volume_i, 'mask': mask_i}
            data_aug_i = np.asarray([aug_pipeline(**data_i) for _ in range(self.aug_factor)])
            vols_aug_i = np.asarray([data_aug_i[j]['image'] for j in range(self.aug_factor)])
            masks_aug_i = np.asarray([data_aug_i[j]['mask'] for j in range(self.aug_factor)])

            for j in range(self.aug_factor):
                tifffile.imwrite(self.aug_volume_path + f'{k}.tif', vols_aug_i[j].transpose(2, 0, 1))
                tifffile.imwrite(self.aug_mask_path + f'{k}.tif', masks_aug_i[j].transpose(2, 0, 1))
                k += 1
        print(f'Number of training volumes: {k}')

    def __len__(self):
        if self.aug_factor is not None:
            return len(os.listdir(self.aug_volume_path))
        else:
            return len(os.listdir(self.split_volume_path))

    def __getitem__(self, idx):
        volname = str(idx) + '.tif'
        midname = os.path.basename(volname)
        if self.aug_factor is not None:
            volume_0 = tifffile.imread(self.aug_volume_path + midname).astype('float32') / 255
            mask_0 = tifffile.imread(self.aug_mask_path + midname).astype('float32') / 255
        else:
            volume_0 = tifffile.imread(self.split_volume_path + midname).astype('float32') / 255
            mask_0 = tifffile.imread(self.split_volume_path + midname).astype('float32') / 255
        volume = torch.from_numpy(volume_0)
        mask = torch.from_numpy(mask_0)
        del volume_0, mask_0
        sample = {'volume': volume, 'mask': mask}
        return sample



import os
import random
import time
from typing import Union, Tuple

import torch.optim as optim
from matplotlib import pyplot as plt
from torch.utils.data import DataLoader, random_split
from torch.utils.tensorboard import SummaryWriter
from tqdm import tqdm

from .data import DataProcess
from .losses import *
from .multi_output_nested_unet import MultiOutputNestedUNet
from ..utils import init_weights, get_device


class Trainer:
    def __init__(self, dataset: DataProcess, num_epochs: int, network=MultiOutputNestedUNet, levels: int = 4,
                 batch_size: int = 4, lr: float = 1e-4, in_channels: int = 1, output_heads: Union[None, dict] = None,
                 n_filter: int = 64, deep_supervision: bool = False,
                 dilation: Union[bool, Tuple[int, int, int, int, int], Tuple[int, int, int, int]] = False,
                 val_split: float = 0.2, save_dir: str = './',
                 save_name: str = 'model.pt', save_iter: bool = False, load_weights: bool = False,
                 device: Union[torch.device, str] = 'auto'):
        if device == 'auto':
            self.device = get_device()
        else:
            self.device = torch.device(device)

        self.model = network(n_filter=n_filter, in_channels=in_channels, output_heads=output_heads, dilation=dilation,
                             deep_supervision=deep_supervision).to(self.device)
        self.model.apply(init_weights)
        self.data = dataset
        self.num_epochs = num_epochs
        self.batch_size = batch_size
        self.levels = levels
        self.lr = lr
        self.best_loss = torch.tensor(float('inf'))
        self.save_iter = save_iter
        self.n_filter = n_filter
        self.dilation = dilation
        self.in_channels = in_channels
        self.output_heads = output_heads

        # Define loss functions for each head
        self.loss_functions = {
            name: self._get_loss_function(config['loss']) for name, config in self.output_heads.items()
        }

        # Define activations for each head
        self.activations = {
            name: config.get('activation', None) for name, config in self.output_heads.items()
        }

        # Define weights for each head
        self.loss_weights = {
            name: config.get('weight', 1.0) for name, config in self.output_heads.items()
        }

        # Split training and validation data
        num_val = int(len(dataset) * val_split)
        num_train = len(dataset) - num_val
        self.dim = dataset.dim_out
        train_data, val_data = random_split(dataset, [num_train, num_val])
        self.train_loader = DataLoader(train_data, batch_size=self.batch_size, pin_memory=True, drop_last=True)
        self.val_loader = DataLoader(val_data, batch_size=self.batch_size, pin_memory=True, drop_last=True)

        self.optimizer = torch.optim.Adam(self.model.parameters(), lr=self.lr)
        self.scheduler = optim.lr_scheduler.ReduceLROnPlateau(self.optimizer, mode='min', patience=5, factor=0.2)

        self.save_dir = save_dir
        self.save_dir_val_result = save_dir + '/val_results/'
        os.makedirs(self.save_dir, exist_ok=True)
        os.makedirs(self.save_dir_val_result, exist_ok=True)
        self.save_name = save_name
        self.params = {
            'optimizer': self.optimizer.state_dict(),
            'lr': self.lr,
            'n_filter': self.n_filter,
            'deep_supervision': deep_supervision,
            'dilation': self.dilation,
            'batch_size': self.batch_size,
            'augmentation': self.data.aug_factor,
            'clip_threshold': self.data.clip_threshold,
            'gauss_noise_lims': self.data.gauss_noise_lims,
            'shot_noise_lims': self.data.shot_noise_lims,
            'brightness_contrast': self.data.brightness_contrast,
            'random_rotate': self.data.random_rotate,
            'in_channels': in_channels,
            'output_heads': output_heads
        }
        if load_weights:
            self.state = torch.load(os.path.join(self.save_dir, self.save_name))
            self.model.load_state_dict(self.state['state_dict'])
            self.epoch_start = self.state['epoch']
        else:
            self.epoch_start = 0

        # Initialize tensorboard
        self.writer = SummaryWriter(log_dir=os.path.join(self.save_dir, 'logs'))

        # Set a fixed random seed for reproducibility
        self.random_seed = 42
        random.seed(self.random_seed)

    @staticmethod
    def _get_loss_function(loss_name):
        if loss_name == 'BCEDiceLoss':
            return BCEDiceLoss()
        elif loss_name == 'DiceLoss':
            return BCEDiceLoss(bce_weight=0, dice_weight=1)
        elif loss_name == 'TverskyLoss':
            return TverskyLoss()
        elif loss_name == 'logcoshTverskyLoss':
            return logcoshTverskyLoss()
        elif loss_name == 'MSELoss':
            return MSELoss()
        elif loss_name == 'MAELoss':
            return MAELoss()
        elif loss_name == 'HuberLoss':
            return HuberLoss()
        elif loss_name == 'DistanceGradientLoss':
            return DistanceGradientLoss()
        elif loss_name == 'WeightedDistanceGradientLoss':
            return WeightedDistanceGradientLoss()
        elif loss_name == 'WeightedVectorFieldLoss':
            return WeightedVectorFieldLoss()
        else:
            raise ValueError(f'Loss "{loss_name}" not defined!')

    @staticmethod
    def _apply_activation(x, activation):
        if activation == 'sigmoid':
            return torch.sigmoid(x)
        elif activation == 'tanh':
            return torch.tanh(x)
        elif activation == 'relu':
            return torch.relu(x)
        elif activation == 'softmax':
            return torch.softmax(x, dim=1)
        return x

    def __iterate(self, epoch, mode):
        if mode == 'train':
            running_loss = 0.0
            for i, batch_i in tqdm(enumerate(self.train_loader), total=len(self.train_loader), unit='batch'):
                x_i = batch_i['image'].to(self.device, non_blocking=True)
                y_i = {key: batch_i[key].to(self.device, non_blocking=True) for key in self.output_heads}

                if x_i.dim() == 3:
                    x_i = x_i.unsqueeze(1)

                # Forward pass
                y_pred = self.model(x_i)

                # Handle deep supervision outputs
                total_loss = 0
                for name, config in self.output_heads.items():
                    target = y_i[name]
                    if target.dim() == 3:
                        target = target.unsqueeze(1)

                    if self.model.deep_supervision:
                        # Calculate loss for each deep supervision output
                        if self.levels == 3:
                            supervision_weights = [0.5, 0.75, 1.0]  # Weights for different supervision levels
                        elif self.levels == 4:
                            supervision_weights = [0.5, 0.75, 0.875, 1.0]
                        else:
                            raise ValueError(f'N = {self.levels} levels not valid. '
                                             f'Choose N=3 or N=4 according to network architecture.')
                        for level, weight in enumerate(supervision_weights, 1):
                            pred_key = f"{name}_{level}"
                            pred = y_pred[pred_key]
                            loss = self.loss_functions[name](pred, target)
                            total_loss += weight * self.loss_weights[name] * loss
                    else:
                        pred = y_pred[name]
                        loss = self.loss_functions[name](pred, target)
                        total_loss += self.loss_weights[name] * loss

                # Backward pass and optimization
                self.optimizer.zero_grad()
                total_loss.backward()
                torch.nn.utils.clip_grad_norm_(self.model.parameters(), max_norm=1.0)
                self.optimizer.step()

                running_loss += total_loss.item()

            avg_train_loss = running_loss / len(self.train_loader)
            self.writer.add_scalar('Loss/train', avg_train_loss, epoch + self.epoch_start)

        elif mode == 'val':
            loss_list = []
            with torch.no_grad():
                for i, batch_i in tqdm(enumerate(self.val_loader), total=len(self.val_loader), unit='batch'):
                    x_i = batch_i['image'].to(self.device, non_blocking=True)
                    y_i = {key: batch_i[key].to(self.device, non_blocking=True) for key in self.output_heads}

                    if x_i.dim() == 3:
                        x_i = x_i.unsqueeze(1)

                    # Forward pass
                    y_pred = self.model(x_i)

                    # Compute combined loss
                    total_loss = 0
                    for name in self.output_heads:
                        target = y_i[name]
                        if target.dim() == 3:
                            target = target.unsqueeze(1)

                        if hasattr(self.model, 'deep_supervision') and self.model.deep_supervision:
                            supervision_weights = [0.5, 0.75, 1.0]
                            for level, weight in enumerate(supervision_weights, 1):
                                pred_key = f"{name}_{level}"
                                pred = self._apply_activation(y_pred[pred_key], self.activations.get(name))
                                loss = self.loss_functions[name](pred, target)
                                total_loss += weight * self.loss_weights[name] * loss
                        else:
                            pred = self._apply_activation(y_pred[name], self.activations.get(name))
                            loss = self.loss_functions[name](pred, target)
                            total_loss += self.loss_weights[name] * loss

                    loss_list.append(total_loss.detach())

            val_loss = torch.stack(loss_list).mean()
            self.writer.add_scalar('Loss/val', val_loss.item(), epoch + self.epoch_start)
            return val_loss

        torch.cuda.empty_cache()

    def plot_images(self, epoch, idx, x_i, y_pred, y_true, output_heads):
        x_i_np = x_i.cpu().numpy().squeeze()

        if hasattr(self.model, 'deep_supervision') and self.model.deep_supervision:
            num_heads = len(output_heads)
            num_levels = self.levels  # Number of deep supervision levels
            fig, axes = plt.subplots(2 + num_levels, num_heads + 1, figsize=(12, 12), dpi=300)

            # Plot input image
            for row in range(2 + num_levels):
                axes[row, 0].imshow(x_i_np, cmap='gray')
                axes[row, 0].set_title('Input' if row == 0 else '')
                axes[row, 0].set_xticks([])
                axes[row, 0].set_yticks([])

            for i, name in enumerate(output_heads):
                cmap = 'viridis' if name in ['length', 'orientation'] else 'gray'

                # Plot final prediction
                final_pred = y_pred[f"{name}"].cpu().numpy().squeeze()
                if len(final_pred.shape) == 3:
                    final_pred = final_pred[0]
                axes[0, i + 1].imshow(final_pred, cmap=cmap)
                axes[0, i + 1].set_title(f'{name} (Final)')
                axes[0, i + 1].set_xticks([])
                axes[0, i + 1].set_yticks([])

                # Plot ground truth
                y_true_np = y_true[name].cpu().numpy().squeeze()
                if len(y_true_np.shape) == 3:
                    y_true_np = y_true_np[0]
                axes[1, i + 1].imshow(y_true_np, cmap=cmap)
                axes[1, i + 1].set_title(f'{name} (True)')
                axes[1, i + 1].set_xticks([])
                axes[1, i + 1].set_yticks([])

                # Plot intermediate supervision outputs
                for level in range(1, num_levels + 1):
                    pred_key = f"{name}_{level}"
                    pred = y_pred[pred_key].cpu().numpy().squeeze()
                    if len(pred.shape) == 3:
                        pred = pred[0]
                    axes[level + 1, i + 1].imshow(pred, cmap=cmap)
                    axes[level + 1, i + 1].set_title(f'Level {level}')
                    axes[level + 1, i + 1].set_xticks([])
                    axes[level + 1, i + 1].set_yticks([])
        else:
            # Original plotting code for non-deep supervision
            num_heads = len(output_heads)
            fig, axes = plt.subplots(2, num_heads + 1, figsize=(12, 8), dpi=300)

            # Plot input image in both rows
            axes[0, 0].imshow(x_i_np, cmap='gray')
            axes[0, 0].set_title('Input')
            axes[0, 0].set_xticks([])
            axes[0, 0].set_yticks([])
            axes[1, 0].imshow(x_i_np, cmap='gray')
            axes[1, 0].set_xticks([])
            axes[1, 0].set_yticks([])

            for i, name in enumerate(output_heads):
                cmap = 'viridis' if name in ['length', 'orientation'] else 'gray'

                # Plot prediction in top row
                pred_np = y_pred[name].cpu().numpy().squeeze()
                if len(pred_np.shape) == 3:
                    pred_np = pred_np[0]
                axes[0, i + 1].imshow(pred_np, cmap=cmap)
                axes[0, i + 1].set_title(f'{name} (Pred)')
                axes[0, i + 1].set_xticks([])
                axes[0, i + 1].set_yticks([])

                # Plot ground truth in bottom row
                true_np = y_true[name].cpu().numpy().squeeze()
                if len(true_np.shape) == 3:
                    true_np = true_np[0]
                axes[1, i + 1].imshow(true_np, cmap=cmap)
                axes[1, i + 1].set_title(f'{name} (True)')
                axes[1, i + 1].set_xticks([])
                axes[1, i + 1].set_yticks([])

        plt.suptitle(f'Epoch {epoch + self.epoch_start}, Sample {idx}')
        plt.tight_layout()
        plt.savefig(self.save_dir_val_result + f'Epoch {epoch + self.epoch_start}, Sample {idx}.png')
        plt.close()

    def log_validation_images(self, epoch, num_images):
        with torch.no_grad():
            all_indices = list(range(len(self.val_loader.dataset)))
            random.seed(self.random_seed)
            selected_indices = random.sample(all_indices, min(num_images, len(all_indices)))

            for idx in selected_indices:
                batch_i = self.val_loader.dataset[idx]

                x_i = batch_i['image'].unsqueeze(0).to(self.device)
                if x_i.dim() == 3:
                    x_i = x_i.unsqueeze(1)

                y_i = {key: batch_i[key].unsqueeze(0).to(self.device) for key in self.output_heads}
                for key in y_i:
                    if y_i[key].dim() == 3:
                        y_i[key] = y_i[key].unsqueeze(1)

                # Forward pass
                y_pred = self.model(x_i)

                if hasattr(self.model, 'deep_supervision') and self.model.deep_supervision:
                    # Handle deep supervision outputs
                    for name in self.output_heads:
                        # Log input image
                        self.writer.add_image(f'Validation/{name}_input_{idx}', x_i[0], epoch)

                        # Log true target image
                        self.writer.add_image(f'Validation/{name}_true_{idx}', y_i[name][0], epoch)

                        # Log predictions for each supervision level
                        for level in range(1, self.levels + 1):
                            pred_key = f"{name}_{level}"
                            pred = self._apply_activation(y_pred[pred_key], self.activations.get(name))
                            self.writer.add_image(f'Validation/{name}_pred_level{level}_{idx}', pred[0],
                                                  epoch + self.epoch_start)
                else:
                    # Original behavior for non-deep supervision
                    y_pred = {name: self._apply_activation(pred, self.activations.get(name))
                              for name, pred in y_pred.items()}

                    for name, pred in y_pred.items():
                        self.writer.add_image(f'Validation/{name}_input_{idx}', x_i[0], epoch + self.epoch_start)
                        self.writer.add_image(f'Validation/{name}_pred_{idx}', pred[0], epoch + self.epoch_start)
                        self.writer.add_image(f'Validation/{name}_true_{idx}', y_i[name][0], epoch + self.epoch_start)

                # Plot images using matplotlib
                self.plot_images(epoch, idx, x_i[0], y_pred, y_i, self.output_heads)

    def start(self):
        for epoch in range(self.num_epochs):
            start_time = time.time()  # Start time of epoch

            # Training phase
            self.__iterate(epoch, 'train')

            # Update state dictionary
            self.state = {
                'epoch': epoch + self.epoch_start,
                'epoch_start': self.epoch_start,
                'best_loss': self.best_loss,
                'state_dict': self.model.state_dict()
            }
            self.state.update(self.params)

            # Validation phase
            with torch.no_grad():
                val_loss = self.__iterate(epoch, 'val')
                self.scheduler.step(val_loss)

            # Calculate time taken for epoch
            end_time = time.time()
            time_taken = round(end_time - start_time, 2)

            print(f"\nEpoch {epoch} completed in {time_taken} seconds.")

            # Check for improvement in validation loss
            if val_loss < self.best_loss:
                print('\nValidation loss improved from %s to %s - saving model state' % (
                    round(self.best_loss.item(), 5), round(val_loss.item(), 5)))
                self.state['best_loss'] = self.best_loss = val_loss
                torch.save(self.state, os.path.join(self.save_dir, self.save_name))

            self.log_validation_images(epoch=epoch, num_images=10)

            # Save model state after each epoch if required
            if self.save_iter:
                torch.save(self.state, os.path.join(self.save_dir, f'model_epoch_{epoch + self.epoch_start}.pt'))


import os
import napari
import numpy as np
import tifffile
import glob
from qtpy.QtWidgets import QPushButton


class ImageAnnotator:
    """
    A class for annotating images using Napari with a single label.

    Parameters
    ----------
    folder_images : str
        Path to the folder containing the images to be annotated.
    output_folder : str
        Path to the folder where the annotated labels will be saved.
    labels_folder : str, optional
        Path to the folder containing preliminary labels.
    label_name : str, optional
        Name of the label layer, by default 'Z-bands'.
    brush_size : int, optional
        Size of the brush for annotation, by default 10.
    threshold : int, optional
        Threshold value for binarizing preliminary labels, by default None.
    """

    def __init__(self, folder_images, output_folder, labels_folder=None, label_name='Z-bands', brush_size=10,
                 threshold=None):
        self.folder_images = folder_images
        self.output_folder = output_folder
        self.labels_folder = labels_folder
        self.label_name = label_name
        self.brush_size = brush_size
        self.threshold = threshold
        self.list_images = glob.glob(os.path.join(self.folder_images, '*.tif'))
        self.current_index = 0

        # Ensure output folder exists
        os.makedirs(self.output_folder, exist_ok=True)

        self.viewer = napari.Viewer()
        self.labels_layer = None

        self.setup_viewer()

    def save_labels(self, labels, filename):
        """
        Save the labels as a uint8 TIFF file.

        Parameters
        ----------
        labels : np.ndarray
            The label data to be saved.
        filename : str
            The path to the file where labels will be saved.
        """
        tifffile.imwrite(filename, (labels.astype(np.uint8) * 255))

    def load_image(self, image_path):
        """
        Load an image into the Napari viewer.

        Parameters
        ----------
        image_path : str
            Path to the image file.

        Returns
        -------
        np.ndarray
            The image data.
        """
        data = tifffile.imread(image_path)
        if len(self.viewer.layers) > 0:
            self.viewer.layers[0].data = data
        else:
            self.viewer.add_image(data)
        return data

    def load_preliminary_labels(self, image_name):
        """
        Load preliminary labels if available and apply threshold if necessary.

        Parameters
        ----------
        image_name : str
            The name of the image file to match the preliminary label file.

        Returns
        -------
        np.ndarray
            The preliminary label data, binarized if a threshold is set.
        """
        if self.labels_folder:
            label_path = os.path.join(self.labels_folder, image_name)
            if os.path.exists(label_path):
                labels = tifffile.imread(label_path)
                if self.threshold is not None:
                    labels = (labels > self.threshold).astype(np.uint8)
                return labels
        return None

    def load_next_image(self):
        """
        Save the current labels and load the next image for annotation.
        """
        # Save current labels
        output_path = os.path.join(self.output_folder, os.path.basename(self.list_images[self.current_index]))
        self.save_labels(self.labels_layer.data, output_path)

        # Increment the index and load the next image
        self.current_index += 1
        if self.current_index < len(self.list_images):
            image_path = self.list_images[self.current_index]
            data = self.load_image(image_path)
            preliminary_labels = self.load_preliminary_labels(os.path.basename(image_path))
            if preliminary_labels is not None:
                self.labels_layer.data = preliminary_labels
            else:
                self.labels_layer.data = np.zeros_like(data)
            self.labels_layer.mode = 'paint'  # Select the brush tool by default
        else:
            print("All images annotated")
            self.viewer.close()

    def setup_viewer(self):
        """
        Set up the Napari viewer with the initial image and annotation tools.
        """
        image_path = self.list_images[self.current_index]
        data = self.load_image(image_path)

        # Load preliminary labels if available
        preliminary_labels = self.load_preliminary_labels(os.path.basename(image_path))
        if preliminary_labels is not None:
            labels_data = preliminary_labels
        else:
            labels_data = np.zeros_like(data)

        # Set brush size and select brush tool by default
        self.labels_layer = self.viewer.add_labels(labels_data, name=self.label_name)
        self.labels_layer.brush_size = self.brush_size  # Adjust the brush size as needed
        self.labels_layer.mode = 'paint'  # Select the brush tool by default

        # Define the save and next function
        def save_and_next(event=None):
            self.load_next_image()

        # Add a button to the viewer
        save_button = QPushButton('Save and Next')
        save_button.clicked.connect(save_and_next)
        self.viewer.window.add_dock_widget(save_button, area='left', name='Save and Next')

        # Add a keybinding to save and next (e.g., pressing 'n')
        @self.viewer.bind_key('n')
        def save_and_next_binding(viewer):
            save_and_next()

        # Run the napari viewer
        napari.run()



import os
from typing import Union

import numpy as np
import tifffile
import torch
from tifffile.tifffile import TiffFile
from tqdm import tqdm as tqdm
from bio_image_unet.progress import ProgressNotifier

from .siam_unet import Siam_UNet
from ..utils import get_device


class Predict:
    """
    Class for prediction of tif-movies with Siamese U-Net


    1) Loading file and preprocess (normalization)
    2) Resizing of images into patches with resize_dim
    3) Prediction with U-Net
    4) Stitching of predicted patches and averaging of overlapping regions

    Parameters
    ----------
    tif_file : str
        Path to input tif stack
    result_name : str
        Path of result file
    model_params : str
        path of u-net parameters (.pt file)
    resize_dim
        Image dimensions for resizing for prediction. If resize_dim=None, the image will not be resized but
        the whole image will be processed by the convolution layers.
    invert : bool
        Invert greyscale of image(s) before prediction
    normalization_mode : str
        Mode for intensity normalization for 3D stacks prior to prediction ('single': each image individually,
        'all': based on histogram of full stack, 'first': based on histogram of first image in stack)
    clip_threshold : Tuple[float, float]
        Clip threshold for image intensity before prediction
    add_tile : int, optional
        Add additional tiles for splitting large images to increase overlap
    normalize_result : bool
        If True, results are normalized to [0, 255]
    device : torch.device or str, optional
        Device to run the pytorch model on, defaults to 'auto', which selects CUDA or MPS if available.
    progress_notifier:
        Wrapper to show tqdm progress notifier in gui
    """

    def __init__(self, tif_file, result_name, model_params, resize_dim=(512, 512), invert=False,
                 normalization_mode='single', clip_threshold=(0.0, 99.98), add_tile=0, normalize_result=False,
                 show_progress=True, device: Union[torch.device, str] = 'auto',
                 progress_notifier: ProgressNotifier = ProgressNotifier.progress_notifier_tqdm()):

        if device == 'auto':
            self.device = get_device()
        else:
            self.device = torch.device(device)

        self.tif_file = tif_file
        self.add_tile = add_tile
        self.invert = invert
        self.normalization_mode = normalization_mode
        self.clip_threshold = clip_threshold
        self.result_name = result_name
        self.normalize_result = normalize_result  # todo to be implemented for Siam-U-Net?
        self.show_progress = show_progress

        # load model
        self.model_params = torch.load(model_params, map_location=self.device)
        self.model = Siam_UNet(n_filter=self.model_params['n_filter'], mode=self.model_params['mode']).to(self.device)
        self.model.load_state_dict(self.model_params['state_dict'])
        self.model.eval()

        # split data into groups of two images
        tif_key = TiffFile(self.tif_file)
        self.tif_len = len(tif_key.pages)
        self.imgs_shape = [self.tif_len, tif_key.pages[0].shape[0], tif_key.pages[0].shape[1]]
        if resize_dim is not None:
            self.resize_dim = resize_dim
        else:
            self.resize_dim = (self.imgs_shape[1], self.imgs_shape[2])

        # create temp folder
        temp_dir = f'temp_{self.tif_file.split("/")[-1]}'

        # taken from split()
        # number of patches in x and y
        self.N_x = int(np.ceil(self.imgs_shape[1] / self.resize_dim[0])) + self.add_tile
        self.N_y = int(np.ceil(self.imgs_shape[2] / self.resize_dim[1])) + self.add_tile
        self.N_per_img = self.N_x * self.N_y
        self.N = self.N_x * self.N_y  # total number of patches

        self.progress_notifier = progress_notifier

        # predict each pair, and save the output of each one as a separate image
        os.makedirs(temp_dir, exist_ok=True)
        print('Predicting data ...') if self.show_progress else None
        with tifffile.TiffWriter(self.result_name, bigtiff=False) as tif:
            if self.show_progress:
                iter = self.progress_notifier.iterator(range(self.tif_len))
            else:
                iter = range(self.tif_len)
            for i, _ in enumerate(iter):
                if i == 0:
                    if self.tif_len == 1:
                        prev_img = tifffile.imread(self.tif_file, key=0)
                    else:
                        prev_img = tifffile.imread(self.tif_file, key=1)
                else:
                    prev_img = current_img
                current_img = tifffile.imread(self.tif_file, key=i)

                img_stack = np.array([prev_img, current_img])
                img_stack = self.__preprocess(img_stack)
                patches = self.__split(img_stack)
                print(f'Patches shape:{patches.shape}') if i == 0 and self.show_progress else None
                result_patches = self.__predict(patches)
                imgs_result = self.__stitch(result_patches)
                tif.write(imgs_result, contiguous=True)

    def __preprocess(self, imgs):
        if len(imgs.shape) == 3:
            if self.normalization_mode == 'single':
                for i, img in enumerate(imgs):
                    img = np.clip(img, a_min=np.nanpercentile(img, self.clip_threshold[0]),
                                  a_max=np.percentile(img, self.clip_threshold[1]))
                    img = img - np.min(img)
                    img = img / np.max(img) * 255
                    if self.invert:
                        img = 255 - img
                    imgs[i] = img
            elif self.normalization_mode == 'first':
                clip_threshold = (np.nanpercentile(imgs[0], self.clip_threshold[0]),
                                  np.percentile(imgs[0], self.clip_threshold[1]))
                imgs = np.clip(imgs, clip_threshold[0], clip_threshold[1])
                imgs = imgs - np.min(imgs)
                imgs = imgs / np.max(imgs) * 255
                if self.invert:
                    imgs = 255 - imgs
            elif self.normalization_mode == 'all':
                clip_threshold = (np.nanpercentile(imgs, self.clip_threshold[0]),
                                  np.percentile(imgs, self.clip_threshold[1]))
                imgs = np.clip(imgs, clip_threshold[0], clip_threshold[1])
                imgs = imgs - np.min(imgs)
                imgs = imgs / np.max(imgs) * 255
                if self.invert:
                    imgs = 255 - imgs
            else:
                raise ValueError(f'normalization_mode {self.normalization_mode} not valid!')
        if len(imgs.shape) == 2:
            imgs = np.clip(imgs, a_min=np.nanpercentile(imgs, self.clip_threshold[0]),
                           a_max=np.percentile(imgs, self.clip_threshold[1]))
            imgs = imgs - np.min(imgs)
            imgs = imgs / np.max(imgs) * 255
            if self.invert:
                imgs = 255 - imgs
        imgs = imgs.astype('uint8')
        return imgs

    def __split(self, imgs):
        # define array for prediction
        patches = np.zeros((self.N, 2, self.resize_dim[0], self.resize_dim[1]), dtype='uint8')

        # zero padding of image if imgs_shape < resize_dim
        if self.imgs_shape[0] > 1:
            if self.imgs_shape[1] < self.resize_dim[0]:  # for x
                imgs = np.pad(imgs, ((0, 0), (0, self.resize_dim[0] - self.imgs_shape[1]), (0, 0)),
                              'constant')
            if self.imgs_shape[2] < self.resize_dim[1]:  # for y
                imgs = np.pad(imgs, ((0, 0), (0, 0), (0, self.resize_dim[1] - self.imgs_shape[2])),
                              'constant')
        elif self.imgs_shape[0] == 1:
            if self.imgs_shape[1] < self.resize_dim[0]:  # for x
                imgs = np.pad(imgs, ((0, self.resize_dim[0] - self.imgs_shape[1]), (0, 0)), 'constant')
            if self.imgs_shape[2] < self.resize_dim[1]:  # for y
                imgs = np.pad(imgs, ((0, 0), (0, self.resize_dim[1] - self.imgs_shape[2])), 'constant')

        # starting indices of patches
        self.X_start = np.linspace(0, self.imgs_shape[1] - self.resize_dim[0], self.N_x).astype('uint16')
        self.Y_start = np.linspace(0, self.imgs_shape[2] - self.resize_dim[1], self.N_y).astype('uint16')

        # split in resize_dim
        n = 0

        i = 1
        for j in range(self.N_x):
            for k in range(self.N_y):
                patches[n, 0, :, :] = imgs[i][self.X_start[j]:self.X_start[j] + self.resize_dim[0],
                                      self.Y_start[k]:self.Y_start[k] + self.resize_dim[1]]
                patches[n, 1, :, :] = imgs[i - 1][self.X_start[j]:self.X_start[j] + self.resize_dim[0],
                                      self.Y_start[k]:self.Y_start[k] + self.resize_dim[1]]
                n += 1
        return patches

    def __predict(self, patches):
        result_patches = np.zeros((patches.shape[0], 1, patches.shape[2], patches.shape[3]), dtype='uint8')
        with torch.no_grad():
            for i, patch_i in enumerate(patches):
                image_patch_i = patch_i[0, :, :]
                prev_image_patch_i = patch_i[1, :, :]

                image_patch_i = torch.from_numpy(image_patch_i.astype('float32') / 255).to(self.device).view(
                    (1, 1, self.resize_dim[0], self.resize_dim[1]))
                prev_image_patch_i = torch.from_numpy(prev_image_patch_i.astype('float32') / 255).to(self.device).view(
                    (1, 1, self.resize_dim[0], self.resize_dim[1]))

                res_i = self.model(image_patch_i, prev_image_patch_i)[0].view(
                    (1, self.resize_dim[0], self.resize_dim[1])).cpu().numpy() * 255
                result_patches[i] = res_i.astype('uint8')
                del patch_i, res_i
        return result_patches

    def __stitch(self, result_patches):
        dtype = result_patches.dtype
        i = 0
        if self.imgs_shape[0] > 1:  # if stack
            stack_result_i = np.zeros((self.N_per_img, np.max((self.resize_dim[0], self.imgs_shape[1])),
                                       np.max((self.resize_dim[1], self.imgs_shape[2]))), dtype=dtype) * np.nan
        elif self.imgs_shape[0] == 1:  # if only one image
            stack_result_i = np.zeros((self.N_per_img, np.max((self.resize_dim[0], self.imgs_shape[1])),
                                       np.max((self.resize_dim[1], self.imgs_shape[2]))), dtype=dtype) * np.nan
        else:
            raise ValueError('Wrong data format!')
        n = 0
        for j in range(self.N_x):
            for k in range(self.N_y):
                stack_result_i[n, self.X_start[j]:self.X_start[j] + self.resize_dim[0],
                self.Y_start[k]:self.Y_start[k] + self.resize_dim[1]] = result_patches[i * self.N_per_img + n, 0, :, :]
                n += 1
        # average overlapping regions
        imgs_result = np.nanmean(stack_result_i, axis=0).astype(dtype)

        # change to input size (if zero padding)
        imgs_result = imgs_result[:self.imgs_shape[1], :self.imgs_shape[2]]

        return imgs_result
